#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/io.h>
#include "m68k.h"
#include "doorsos.h"
#include "objimage.h"
#include "link.h"
#include "tiformat.h"

char progstub [] = {
      0x4A,0x78,0x00,0x30,0x66,0x04,0x58,0x8F,
      0x4E,0x75,0x20,0x78,0x00,0x34,0x4E,0xD0};
int prog_stubsize = 16;

char libstub [] = {
      0x48,0xE7,0xFF,0xFE,0x48,0x7A,0x00,0x1A,
      0x28,0x78,0x00,0xC8,0xD9,0xFC,0x00,0x00,
      0x03,0x98,0x28,0x54,0x4E,0x94,0x58,0x8F,
      0x4C,0xDF,0x7F,0xFF,0x58,0x8F,0x4E,0x75,

      0x4C,0x69,0x62,0x72,0x61,0x72,0x69,0x65,
      0x73,0x20,0x61,0x72,0x65,0x20,0x6E,0x6F,
      0x74,0x20,0x65,0x78,0x65,0x63,0x75,0x74,
      0x61,0x62,0x6C,0x65,0x00,0x00};
int lib_stubsize = 62;

int error = 0;

extern char* project;
extern BOOL quiet;

ASM *binary;
ROM *rom;
RAM *ram;
BSStable *bss;
struct mod_index *indexes, *indexes16;
reloc *rel;
ushort hdrsize;
import *im;
int i,j,k,l;
unsigned char *buf;

int write_TIfile(TIfile *file,char *name) // name w/o extension
{
 char projet[13];
 FILE* fout;
 ushort* usp;
 char* RAMfile; // image du fichier final en RAM
 int ti89=0;
 int ti92plus=0;

 binary = file->core->raw;
 if (!(binary->flags & 0x03))
 {
  fprintf(stderr,"Error: Neither _ti89 nor _ti92plus is defined\n");
  return 0;
 }
 if (!binary->main_offset && strncmp("68kL",binary->sign,4) && !binary->nostub)
 {
  fprintf(stderr,"Error: Fatal: no entry point '_main' defined\n");
  return 0;
 }

 if (binary->flags & 0x01)
  ti92plus = 1;

 if (binary->flags & 0x02)
  ti89 = 1;

 buf = alloc(file->size);
 RAMfile = buf;
 memset(buf,0,file->size); // met tout  0..

 memcpy(buf,file,0x56); // copie le header
 buf+=0x56;

 buf_ushort(file->core->size); // taille



 if (binary->nostub)
 {
  if (binary->bss_offset ||
      binary->extraRAM_offset ||
      binary->importtable->nblibs ||
      binary->romtable->flag ||
      binary->ramtable->flag)
  {
   fputs("Error: Fatal: _nostub is not compatible with BSS, ROM CALLS,RAM CALLS, libraries\n",stderr);
   return 0;
  }
  memcpy(buf,binary->code,binary->codesize-1);
  buf+= binary->codesize-1;
  for (i=0;i<binary->reloctable->nboffsets;i++)
  {
   ushort ofs = binary->reloctable->offsets[i];
   *(ushort*)(buf) = *(ushort*)(RAMfile+ofs+2+0x58);
   buf+=2;
   buf_ushort(ofs);
  }
  *buf++ = 0xF3;
 }
 else
 {
  usp = &binary->bsr;
  rusp(usp++);  // binary->bsr
  rusp(usp);    // binary->bsr_offset
  usp += 3;     // incremente..
  for (i=0;i<8;i++)
   rusp(usp++);

  memcpy(buf,&binary->bsr,26); // le header du header du prog :)
  buf +=26;

                 // S'OCCUPE DES LIBS //

  im = binary->importtable;
  buf_ushort(im->nblibs);
  for (i=0;i<im->nblibs;i++)
  {
   memcpy(buf,im->libs_names[i],10);
   buf += 10;
  }
  for (i=0;i<im->nblibs;i++)
  {
   buf_ushort(im->lib_reloc[i]->nbfunc);
   for (j=0;j<im->lib_reloc[i]->nbfunc+1;j++)
   {
    reloc_sym* rsym = im->lib_reloc[i]->table[j];
    buf_ushort(rsym->funcindex);

    for (k=0;k<rsym->reltable.nboffsets;k++)
     buf_ushort(rsym->reltable.offsets[k]);

    buf_ushort(0); // 0 de fin de table
   }
  }

                 // ROM TABLE //

  rom = binary->romtable;
  buf_ushort(rom->flag);
  if (rom->flag)
  {
   buf_ushort(rom->ROM_reloc->nbfunc); // nb de func utilises
   for (i=0;i<rom->ROM_reloc->nbfunc+1;i++)
   {
    reloc_sym* rsym = rom->ROM_reloc->table[i];
    buf_ushort(rsym->funcindex);
    for (k=0;k<rsym->reltable.nboffsets;k++)
      buf_ushort(rsym->reltable.offsets[k]);
    buf_ushort(0); // 0 de fin de table
   }
  }

                 // RAM TABLE //

  ram = binary->ramtable;
  buf_ushort(ram->flag);
  if (ram->flag)
  {
   i = (ram->RAM32_reloc) ? ram->RAM32_reloc->nbfunc+1 : 0;
   i += (ram->RAM16_reloc) ? ram->RAM16_reloc->nbfunc+1 : 0;
   i += - 1;

   buf_ushort(i); // nb de func utilises
   if (ram->RAM32_reloc)
   {
    for (i=0;i<ram->RAM32_reloc->nbfunc+1;i++)
    {
     reloc_sym* rsym = ram->RAM32_reloc->table[i];
     buf_ushort(rsym->funcindex);
     for (k=0;k<rsym->reltable.nboffsets;k++)
       buf_ushort(rsym->reltable.offsets[k]);
     buf_ushort(0); // 0 de fin de table
    }
   }
   if (ram->RAM16_reloc)
   {
    for (i=0;i<ram->RAM16_reloc->nbfunc+1;i++)
    {
     reloc_sym* rsym = ram->RAM16_reloc->table[i];
     buf_ushort(rsym->funcindex|0x8000);
     for (k=0;k<rsym->reltable.nboffsets;k++)
       buf_ushort(rsym->reltable.offsets[k]);
     buf_ushort(0); // 0 de fin de table
    }
   }
  }

                 // RELOC TABLE //

  rel = binary->reloctable;
  for (i=0;i<rel->nboffsets;i++)
   buf_ushort(rel->offsets[i]);
  buf_ushort(0);

                 // BSS TABLE //

  bss = binary->bss;
  if (bss->size)
  {
   buf_ulong(bss->size);
   rel = bss->BSS_reloc;
   for (i=0;i<rel->nboffsets;i++)
    buf_ushort(rel->offsets[i]);
   buf_ushort(0);
  }

                 // EXPORT TABLE //

  rel = binary->exporttable;
  if (rel->nboffsets)
  {
   buf_ushort(rel->nboffsets);
   for (i=0;i<rel->nboffsets;i++)
    buf_ushort(rel->offsets[i]);
   buf_ushort(0);
  }
                // STUB //

  memcpy(buf,binary->stub,binary->stubsize);
  buf += binary->stubsize;

                // CODE //

  memcpy(buf,binary->code,binary->codesize);
  buf += binary->codesize;
 }
                // CHECKSUM //
 buf = RAMfile;
 j = file->size - 0x58;
 k = 0; // k sert de checksum
 for (i=0x56;i<file->size-2;i++)
  k += buf[i];

 buf = RAMfile + file->size - 2;
 buf_ushort(rus(k)); // met le checksum


 if (ti89) // si on fait un 89
 {
  sprintf(projet,"%s.89z",name);
  strncpy(RAMfile,"**TI89**",8);

  if ((fout = fopen(projet,"wb")) == 0)
  {
   fprintf(stderr,"Error: can't open destination file: %s\n",projet);
   return 0;
  }
  if (fwrite(RAMfile,1,file->size,fout) == -1)
  {
   fclose(fout);
   fprintf(stderr,"Error: can't write destination file: %s\n",projet);
   return 0;
  }
  fclose(fout);
  if(!quiet)
  {
   printf("Linked %s  Size: %d bytes\n",projet,file->core->size+2);
  }
 }

 if (ti92plus) // si on fait un 92+
 {
  sprintf(projet,"%s.9xz",name);
  strncpy(RAMfile,"**TI92P*",8);

  if ((fout = fopen(projet,"wb")) == 0)
  {
   fprintf(stderr,"Error: can't open destination file: %s\n",projet);
   return 0;
  }
  if (fwrite(RAMfile,1,file->size,fout) == -1)
  {
   fclose(fout);
   fprintf(stderr,"Error: can't write destination file: %s\n",projet);
   return 0;
  }
  fclose(fout);
  if(!quiet)
  {
   printf("Linked %s  Size: %d bytes\n",projet,file->core->size+2);
  }
 }
 return 1;
}

void buf_ulong(ulong dword)
{
 *(ulong*)buf = rul(dword);
 buf+=4;
}

void buf_ushort(ushort word)
{
 *(ushort*)buf = rus(word);
 buf+=2;
}

/*******************
* Create a TI file *
*******************/
TIfile* object_to_TI(objimage* obj)
{
 int nb_names = 0;
 char folder[] = {'m','a','i','n',0,0,0,0};
 int i,j,k,l;
 TIfile* compiled = alloc(sizeof(TIfile));
 int offs[2];
 offs[0] = 0;
 offs[1] = obj->sections[CODE].size;

                // Alloue les bases du fichier

 compiled->unused1 = 1;
 memcpy(compiled->folder,folder,8);
 memset(compiled->desc, 0, 40+6+8+4+4+6);
 compiled->unused2[0]=1;
 compiled->unused2[2]=0x52;

 strcpy(compiled->file,project);
 compiled->type = 0x21;
 compiled->unused3[0] = 0xA5;
 compiled->unused3[1] = 0x5A;

 compiled->core = alloc(sizeof(filecore));
 compiled->core->raw = alloc(sizeof(ASM));
 binary = compiled->core->raw;
 binary->codesize = obj->sections[0].size + obj->sections[1].size+3; // taille de .text + .data + 3 bytes:$0000F3
 binary->code = alloc(binary->codesize);
 memcpy(binary->code,
        obj->sections[0].raw,
        obj->sections[0].size);   // copie le .text

 memcpy(binary->code+obj->sections[0].size,
        obj->sections[1].raw,
        obj->sections[1].size); // copie le .data

                // Set file type
 binary->code[binary->codesize-3] = 0;
 binary->code[binary->codesize-2] = 0;
 binary->code[binary->codesize-1] = 0xF3;

                // Initialise

 binary->bsr = 0x6100;     // 0x6100 -> 'bsr'
 binary->bsr_offset = 0;
 binary->relcnt = 0;
 strncpy(binary->sign,"68kP",4);
 binary->stubsize = prog_stubsize;
 binary->stub = progstub;
 binary->nostub = 0;


 binary->comment_offset = 0xFFFF;
 binary->main_offset = 0xFFFF;
 binary->exit_offset = 0xFFFF;
 binary->flags = 0;
 binary->bss_hdl = 0;
 binary->bss_offset = 0;
 binary->export_offset = 0;
 binary->extraRAM_offset = 0;

                // s'occupe de l'import table
 binary->importtable = alloc(sizeof(import));
 im = binary->importtable;
 im->libs_names = 0; // pointeur  0
                // remplit les noms des libs
 for (i=0;i<2;i++)
 {
  for (j=0;j<obj->sections[i].rel.nb_unr;j++)
  {
   char name[9];

   char* str = obj->sections[i].rel.unr[j]->sym->name;
   k = strlen(str);
   if (!is_lib_name(str)) continue; //si c pas une lib
   if ( k<7 || k>14 ) continue;
   if (strncmp("__",str+k-6,2)) continue;
   strncpy(name,str,k-6);  // copie le nom de la lib
   name[k-6] = 0;          // termine la chaine
   for (k=0;k<nb_names;k++)
   {
    if (!strcmp((im->libs_names[k]),name)) // si le nom existe
     break;
   }
   if (k==nb_names)             // si le nom n'a pas t trouv
   {
    nb_names++;                 // nblibs ++
    im->libs_names = re_alloc(im->libs_names,4*nb_names); // realloue la table des ptrs vers les noms de libs
    im->libs_names[nb_names-1] = alloc(10); // alloue 10 octets pour le nom
    memset(im->libs_names[nb_names-1],0,10); // met a 0 les 10 octets
    strcpy(im->libs_names[nb_names-1],name); // copie le nom
   }
  }
 }
 binary->importtable->nblibs = nb_names;

                   // ALLOUE LA TABLE DE RELOC POUR LES LIBS //

 binary->importtable->lib_reloc = alloc(nb_names*sizeof(reloc_module*));
 for (j=0;j<nb_names;j++)
  binary->importtable->lib_reloc[j] = alloc(sizeof(reloc_module));

 for (j=0;j<nb_names;j++)       // pour ttes les libs
 {
  char name[10];
  strcpy(name,im->libs_names[j]);
  k = strlen(name);
  name[k] = '_'; name[k+1] = 0;    // ajoute '_' au nom de la lib
  indexes = nb_index_used(obj,name);
  im->lib_reloc[j]->nbfunc = indexes->nbindex-1;
  im->lib_reloc[j]->table = alloc(4*indexes->nbindex);
  for (l=0;l<indexes->nbindex;l++)
  {
   im->lib_reloc[j]->table[l] = alloc(sizeof(reloc_sym));
   im->lib_reloc[j]->table[l]->funcindex=indexes->indexes[l];
   im->lib_reloc[j]->table[l]->reltable.nboffsets = 0;
   im->lib_reloc[j]->table[l]->reltable.offsets = 0;
  }
  free (indexes->indexes);
  free (indexes);
 }

                // Remplit la lib_reloc //
 for (i=0;i<2;i++)
 {
  for (j=0;j<obj->sections[i].rel.nb_unr;j++)
  {
   reloc_sym* rel;
   char* str = obj->sections[i].rel.unr[j]->sym->name;
   if (!is_lib_name(str)) continue;
   rel = lib_rel_tbl(str,im);  // renvoie le ptr vers la bonne rel table dans
   k = ++rel->reltable.nboffsets;
   rel->reltable.offsets = re_alloc(rel->reltable.offsets,sizeof(ushort)*k);
   rel->reltable.offsets[k-1] = obj->sections[i].rel.unr[j]->offset+
                                offs[i];
  }
 }

                // S'OCCUPE DE LA ROM TABLE //

 binary->romtable = alloc(sizeof(ROM));
 rom = binary->romtable;
 indexes = nb_index_used(obj,"_ROM_CALL");
 i = indexes->nbindex;
 if (i == 0)
 {
  rom->flag = 0;
  rom->ROM_reloc = 0;
 }
 else
 {
  rom->flag = 1;
  rom->ROM_reloc = alloc(sizeof(reloc_module));
  rom->ROM_reloc->nbfunc = i-1;
  rom->ROM_reloc->table = alloc(4*i);
  for (j=0;j<i;j++)
  {
   rom->ROM_reloc->table[j] = alloc(sizeof(reloc_sym));
   rom->ROM_reloc->table[j]->funcindex = indexes->indexes[j];
   rom->ROM_reloc->table[j]->reltable.nboffsets = 0;
   rom->ROM_reloc->table[j]->reltable.offsets = 0;
  }
 }
 free (indexes->indexes);
 free (indexes);

 if (rom->flag)
 {
  for (i=0;i<2;i++)
  {
   for (j=0;j<obj->sections[i].rel.nb_unr;j++)
   {
    reloc_sym* rel;
    char* str = obj->sections[i].rel.unr[j]->sym->name;
    if (!strncmp("_ROM_CALL_",str,10))
    {
     rel = sym_rel_tbl(str,rom->ROM_reloc);  // renvoie le ptr vers la bonne rel table dans
     k = ++rel->reltable.nboffsets;
     rel->reltable.offsets = re_alloc(rel->reltable.offsets,sizeof(ushort)*k);
     rel->reltable.offsets[k-1] = obj->sections[i].rel.unr[j]->offset+
                                  offs[i];
    }
   }
  }
 }

                // S'OCCUPE DE LA RAM TABLE //

 binary->ramtable = alloc(sizeof(RAM));
 ram = binary->ramtable;
 indexes = nb_index_used(obj,"_RAM_CALL");
 indexes16 = nb_index16_used(obj,"_RAM_CALL");
 i = indexes->nbindex + indexes16->nbindex;
 if (i == 0)
 {
  ram->flag = 0;
  ram->RAM32_reloc = 0;
  ram->RAM16_reloc = 0;
 }
 else
 {
  ram->flag = 1;

  if (indexes->nbindex)
  {
   ram->RAM32_reloc = alloc(sizeof(reloc_module));
   ram->RAM32_reloc->nbfunc = indexes->nbindex-1;
   ram->RAM32_reloc->table = alloc(4*indexes->nbindex);
   for (j=0;j<indexes->nbindex;j++)
   {
    ram->RAM32_reloc->table[j] = alloc(sizeof(reloc_sym));
    ram->RAM32_reloc->table[j]->funcindex = indexes->indexes[j];
    ram->RAM32_reloc->table[j]->reltable.nboffsets = 0;
    ram->RAM32_reloc->table[j]->reltable.offsets = 0;
   }
  }
  else
   ram->RAM32_reloc = 0;

  if (indexes16->nbindex)
  {
   ram->RAM16_reloc = alloc(sizeof(reloc_module));
   ram->RAM16_reloc->nbfunc = indexes16->nbindex-1;
   ram->RAM16_reloc->table = alloc(4*indexes16->nbindex);
   for (j=0;j<indexes16->nbindex;j++)
   {
    ram->RAM16_reloc->table[j] = alloc(sizeof(reloc_sym));
    ram->RAM16_reloc->table[j]->funcindex = indexes16->indexes[j];
    ram->RAM16_reloc->table[j]->reltable.nboffsets = 0;
    ram->RAM16_reloc->table[j]->reltable.offsets = 0;
   }
  }
  else
   ram->RAM16_reloc = 0;
 }
 free (indexes->indexes);
 free (indexes);
 free (indexes16->indexes);
 free (indexes16);

 if (ram->flag)
 {
  if (ram->RAM32_reloc)
  {
   for (i=0;i<2;i++)
   {
    for (j=0;j<obj->sections[i].rel.nb_unr;j++)
    {
     reloc_sym* rel;
     char* str = obj->sections[i].rel.unr[j]->sym->name;
     if (obj->sections[i].rel.unr[j]->type)
      continue;
     if (!strncmp("_RAM_CALL_",str,10))
     {
      rel = sym_rel_tbl(str,ram->RAM32_reloc);  // renvoie le ptr vers la bonne rel table
      k = ++rel->reltable.nboffsets;
      rel->reltable.offsets = re_alloc(rel->reltable.offsets,sizeof(ushort)*k);
      rel->reltable.offsets[k-1] = obj->sections[i].rel.unr[j]->offset+
                                   offs[i];
     }
    }
   }
  }
  if (ram->RAM16_reloc)
  {
   for (i=0;i<2;i++)
   {
    for (j=0;j<obj->sections[i].rel.nb_unr;j++)
    {
     reloc_sym* rel;
     char* str = obj->sections[i].rel.unr[j]->sym->name;
     if (!obj->sections[i].rel.unr[j]->type)
      continue;
     if (!strncmp("_RAM_CALL_",str,10))
     {
      rel = sym_rel_tbl(str,ram->RAM16_reloc);  // renvoie le ptr vers la bonne rel table
      k = ++rel->reltable.nboffsets;
      rel->reltable.offsets = re_alloc(rel->reltable.offsets,sizeof(ushort)*k);
      rel->reltable.offsets[k-1] = obj->sections[i].rel.unr[j]->offset+
                                   offs[i];
     }
    }
   }
  }
 }

                  // S'OCCUPE DE LA RELOC TABLE //

 binary->reloctable = alloc(sizeof(reloc));
 rel = binary->reloctable;
 rel->nboffsets = 0;
 rel->offsets = 0;
 for (i=CODE;i<BSS;i++)
 {
  for (j=0;j<obj->sections[i].rel.nb_prv;j++)
  {
   k = ++rel->nboffsets;
   rel->offsets = re_alloc(rel->offsets,k*sizeof(ushort));
   rel->offsets[k-1] = obj->sections[i].rel.prv[j]->offset
                       + offs[i];
  }
 }

                  // S'OCCUPE DE LA BSS TABLE

 binary->bss = alloc(sizeof(BSStable));
 bss = binary->bss;
 bss->size = obj->sections[2].size;
 bss->BSS_reloc = alloc(sizeof(reloc));
 rel = bss->BSS_reloc;
 rel->nboffsets = 0;
 rel->offsets = 0;
 for (i=0;i<2;i++)
 {
  for (j=0;j<obj->sections[i].rel.nb_bss;j++)
  {
   k = ++rel->nboffsets;
   rel->offsets = re_alloc(rel->offsets,k*sizeof(ushort));
   rel->offsets[k-1] = obj->sections[i].rel.bss[j]->offset;
  }
 }



                  // S'OCCUPE DE L'EXPORT TABLE,de main, etc etc.. //

 binary->exporttable = alloc(sizeof(reloc));
 rel = binary->exporttable;
 rel->nboffsets = 0;
 rel->offsets = 0;
 for (i=0;i<obj->symbs.nb_pub;i++)
 {
  spublic* sym = obj->symbs.pub_sym[i];
  if (!strcmp(sym->name,"_main"))
  {
   binary->main_offset = sym->offset;
   continue;
  }

  if (!strcmp(sym->name,"_comment"))
  {
   binary->comment_offset = sym->offset;
   continue;
  }

  if (!strcmp(sym->name,"_exit"))
  {
   binary->exit_offset = sym->offset;
   continue;
  }

  if (!strcmp(sym->name,"_ti89"))
  {
   binary->flags |= 0x2;
   continue;
  }

  if (!strcmp(sym->name,"_ti92plus"))
  {
   binary->flags |= 0x1;
   continue;
  }

  if (!strcmp(sym->name,"_library"))
  {
   strncpy(binary->sign,"68kL",4);
   binary->stubsize = lib_stubsize;
   binary->stub = libstub;
   continue;
  }

  if (!strcmp(sym->name,"_prosit"))
  {
   binary->stubsize = 0; // pas de stub
   binary->bsr = 0x4E75; // rts
   binary->bsr_offset = sym->offset+2;
   continue;
  }

  if (!strcmp(sym->name,"_nostub"))
  {
   binary->nostub = 1;
   continue;
  }

  if (is_lib_name(sym->name))
  {
   if (!strncmp(project,sym->name,strlen(project)))
   {
    ushort indx = 1+strtol(sym->name+strlen(project)+2,NULL,16); // indice de l'export
    if (indx > rel->nboffsets)
    {
     rel->nboffsets = indx;             // A AMELIORER ICI
     rel->offsets = re_alloc(rel->offsets,indx*sizeof(ushort));
    }
    rel->offsets[indx-1] = sym->offset;
    continue;
   }
  }
 }
                // VERIFIE QU'IL N'Y A PAS DE SYMBOLES UNRELOCATED //
 error = 0;
 for (i=0;i<2;i++)
 {
  for (j=0;j<obj->sections[i].rel.nb_unr;j++)
  {
   char* str = obj->sections[i].rel.unr[j]->sym->name;
   if (obj->sections[i].rel.unr[j]->sym->enabled == 0)
    continue;
   if (is_lib_name(str))
    continue;
   if (!strncmp("_ROM_CALL_",str,10))
    continue;
   if (!strncmp("_RAM_CALL_",str,10))
    continue;
   fprintf(stderr,"Error: undefined reference to '%s'\n",str);
   error++;
  }
 }
 if (error)
  return 0;

                // CALCUL DE LA TAILLE DU HEADER //

 hdrsize = 26
         + 2+im->nblibs*12 // libs names & nblibs
         + 2               // ROM flag
         + 2               // RAM flag
         + (binary->reloctable->nboffsets+1)*2 // Reloc Table
         + binary->stubsize;

 for (i=0;i<im->nblibs;i++)
 {
  for (j=0;j<im->lib_reloc[i]->nbfunc+1;j++)
  {
   hdrsize += 2;  // index of the function used
   hdrsize += (im->lib_reloc[i]->table[j]->reltable.nboffsets+1)*2; // Ajoute la taille des reloc de libs
  }
 }

 if (rom->flag)
 {
  hdrsize += 2; // nb index used
  for (i=0;i<rom->ROM_reloc->nbfunc+1;i++)
  {
   hdrsize += 2; // index of the function used
   hdrsize += (rom->ROM_reloc->table[i]->reltable.nboffsets+1)*2; // reloc ROM
  }
 }

 if (ram->flag)
 {
  hdrsize += 2; // nb index used
  if (ram->RAM32_reloc)
  {
   for (i=0;i<ram->RAM32_reloc->nbfunc+1;i++)
   {
    hdrsize += 2; // index of the function used
    hdrsize += (ram->RAM32_reloc->table[i]->reltable.nboffsets+1)*2; // reloc ROM
   }
  }
  if (ram->RAM16_reloc)
  {
   for (i=0;i<ram->RAM16_reloc->nbfunc+1;i++)
   {
    hdrsize += 2; // index of the function used
    hdrsize += (ram->RAM16_reloc->table[i]->reltable.nboffsets+1)*2; // reloc ROM
   }
  }
 }

 if (bss->size)
 {
  binary->bss_offset = hdrsize - binary->stubsize;
  hdrsize += 4; // taille de la section BSS
  hdrsize += (bss->BSS_reloc->nboffsets+1)*2; // relocs BSS
 }

 if (binary->exporttable->nboffsets)
 {
  binary->export_offset = hdrsize - binary->stubsize;
  hdrsize += 2; // nb de func exportes
  hdrsize += (binary->exporttable->nboffsets+1)*2;
 }

 binary->bsr_offset += hdrsize - binary->stubsize - 2;

                    // Voila ! Hdrsize est calcul ! //
                    // Maintenant, va pour l'ajouter partout :-( //

 if (binary->comment_offset == 0xFFFF)
  binary->comment_offset = 0;
 else
  binary->comment_offset += hdrsize;

 if (binary->exit_offset == 0xFFFF)
  binary->exit_offset = 0;
 else
  binary->exit_offset += hdrsize;

 if (binary->main_offset == 0xFFFF)
  binary->main_offset = 0;
 else
  binary->main_offset += hdrsize;


 for (i=0;i<im->nblibs;i++)
 {
  for (j=0;j<im->lib_reloc[i]->nbfunc+1;j++)
  {
   for (k=0;k<im->lib_reloc[i]->table[j]->reltable.nboffsets;k++)
    im->lib_reloc[i]->table[j]->reltable.offsets[k] += hdrsize; // ajoute la taille du header
  }
 }

 if (rom->flag)
 {
  for (i=0;i<rom->ROM_reloc->nbfunc+1;i++)
  {
   for (j=0;j<rom->ROM_reloc->table[i]->reltable.nboffsets;j++)
      rom->ROM_reloc->table[i]->reltable.offsets[j] += hdrsize;
  }
 }

 if (ram->flag)
 {
  if (ram->RAM32_reloc)
  {
   for (i=0;i<ram->RAM32_reloc->nbfunc+1;i++)
   {
    for (j=0;j<ram->RAM32_reloc->table[i]->reltable.nboffsets;j++)
     ram->RAM32_reloc->table[i]->reltable.offsets[j] += hdrsize;
   }
  }
  if (ram->RAM16_reloc)
  {
   for (i=0;i<ram->RAM16_reloc->nbfunc+1;i++)
   {
    for (j=0;j<ram->RAM16_reloc->table[i]->reltable.nboffsets;j++)
     ram->RAM16_reloc->table[i]->reltable.offsets[j] += hdrsize;
   }
  }
 }
 if (!binary->nostub)
 {
  for (i=0;i<binary->reloctable->nboffsets;i++)
  {
   ulong l = rul(*(ulong*)(binary->code+binary->reloctable->offsets[i]));
   *(ulong*)(binary->code+binary->reloctable->offsets[i]) = rul(hdrsize+l);
   binary->reloctable->offsets[i] += hdrsize;
  }
 }

 for (i=0;i<bss->BSS_reloc->nboffsets;i++)
  binary->bss->BSS_reloc->offsets[i] += hdrsize;

 for (i=0;i<binary->exporttable->nboffsets;i++)
  binary->exporttable->offsets[i] += hdrsize;

                     // Voila !

 if (!binary->nostub)
  compiled->core->size = hdrsize+binary->codesize; // header + prog
 else
  compiled->core->size = binary->codesize+(4*binary->reloctable->nboffsets); // prog + (reloc table=4*nboffsets + $00)

 compiled->size = 0x58 + compiled->core->size + 2; //TI header + prog + csum

 return compiled;
}

reloc_sym* lib_rel_tbl(char* str, import* im)
{
 int i,j;
 char symname[10];
 i = strlen(str) - 6;
 strncpy(symname,str,i);
 symname[i] = 0;        // termine la chaine
 for (j=0;j<im->nblibs;j++)
  if (!strcmp(symname,im->libs_names[j])) break; // si le nom est trouv, j = index de la lib
 return sym_rel_tbl(str,im->lib_reloc[j]);
}

reloc_sym* sym_rel_tbl(char* str,reloc_module* rel)
{
 int i;
 ushort index;
 char name[20];
 char* ptr;
 strcpy(name,str);
 ptr = name + strlen(name);
 while (*ptr != '_')
  ptr--;
 ptr++; // place apres le '_'

 index = strtol(ptr,NULL,16);
 for (i=0;i<rel->nbfunc+1;i++)
  if (rel->table[i]->funcindex == index) return rel->table[i];

 return 0; // si on a rien trouv...
}

struct mod_index* nb_index_used(objimage* obj, char* symname)
{
 int i,j,k;
 struct mod_index* indexes = alloc(sizeof(struct mod_index));
 indexes->nbindex = 0;
 indexes->indexes = 0;
 for (i=0;i<2;i++)
 {
  for (j=0;j<obj->sections[i].rel.nb_unr;j++)
  {
   if (!strncmp(obj->sections[i].rel.unr[j]->sym->name,symname,strlen(symname)))
   {
    ushort funcnb;
    funcnb = strtol(obj->sections[i].rel.unr[j]->sym->name+strlen(symname)+1,NULL,16);
    if (obj->sections[i].rel.unr[j]->type) // si rel 16 bits: bye
     continue;

    for (k=0;k<indexes->nbindex;k++)
    {
     if (indexes->indexes[k] == funcnb)
      break;
    }
    if (k==indexes->nbindex)
    {
     indexes->nbindex++;
     indexes->indexes = re_alloc(indexes->indexes,sizeof(ushort)*indexes->nbindex);
     indexes->indexes[k] = funcnb;
    }
   }
  }
 }
 return indexes;
}

struct mod_index* nb_index16_used(objimage* obj, char* symname)
{
 int i,j,k;
 struct mod_index* indexes = alloc(sizeof(struct mod_index));
 indexes->nbindex = 0;
 indexes->indexes = 0;
 for (i=0;i<2;i++)
 {
  for (j=0;j<obj->sections[i].rel.nb_unr;j++)
  {
   if (!strncmp(obj->sections[i].rel.unr[j]->sym->name,symname,strlen(symname)))
   {
    ushort funcnb;
    funcnb = strtol(obj->sections[i].rel.unr[j]->sym->name+strlen(symname)+1,NULL,16);
    if (!obj->sections[i].rel.unr[j]->type) // si rel 16 bits: bye
     continue;
    for (k=0;k<indexes->nbindex;k++)
    {
     if (indexes->indexes[k] == funcnb)
      break;
    }
    if (k==indexes->nbindex)
    {
     indexes->nbindex++;
     indexes->indexes = re_alloc(indexes->indexes,sizeof(ushort)*indexes->nbindex);
     indexes->indexes[k] = funcnb;
    }
   }
  }
 }
 return indexes;
}

void delete_TIfile(TIfile* file)
{
 int i,j;
 import* im = file->core->raw->importtable;
 ROM* rom = file->core->raw->romtable;
 RAM* ram = file->core->raw->ramtable;

        // free la import table

 free (im->libs_names);
 for (i=0;i<im->nblibs;i++)
 {
  for (j=0;j<im->lib_reloc[i]->nbfunc+1;j++)
  {
   free (im->lib_reloc[i]->table[j]->reltable.offsets);
   free (im->lib_reloc[i]->table[j]);
  }
  free (im->lib_reloc[i]);
 }
 free (im->lib_reloc);
 free (im);

       // free la ROM table

 if (rom->flag)
 {
  for (i=0;i<rom->ROM_reloc->nbfunc;i++)
  {
   free (rom->ROM_reloc->table[i]->reltable.offsets);
   free (rom->ROM_reloc->table[i]);
  }
 }
 free (rom->ROM_reloc);
 free (rom);

       // free la RAM table

 if (ram->flag)
 {
  if (ram->RAM32_reloc)
  {
   for (i=0;i<ram->RAM32_reloc->nbfunc;i++)
   {
    free (ram->RAM32_reloc->table[i]->reltable.offsets);
    free (ram->RAM32_reloc->table[i]);
   }
  }
  if (ram->RAM16_reloc)
  {
   for (i=0;i<ram->RAM16_reloc->nbfunc;i++)
   {
    free (ram->RAM16_reloc->table[i]->reltable.offsets);
    free (ram->RAM16_reloc->table[i]);
   }
  }
 }
 free (ram->RAM32_reloc);
 free (ram->RAM16_reloc);
 free (ram);

       // free la reloc table

 free (file->core->raw->reloctable->offsets);
 free (file->core->raw->reloctable);

       // free la BSS table

 free (file->core->raw->bss->BSS_reloc->offsets);
 free (file->core->raw->bss->BSS_reloc);
 free (file->core->raw->bss);

       // free l'export table

 free (file->core->raw->exporttable->offsets);
 free (file->core->raw->exporttable);

 free (file->core->raw->code);
 free (file->core->raw);
 free (file->core);
 free (file);

}

